import logging

from kensu.utils.helpers import singleton


@singleton
class KensuProvider(object):
    default = None  # type: Kensu

    def instance(self):
        # type: () -> Kensu
        return self.default

    def setKensu(self, kensu):
        self.default = kensu

    # FIXME: probably we don't need get_explicit_code_version_fn & get_code_version_fn anymore. but keeping it for backwards-compat for now...
    @staticmethod
    def initKensu(api_url=None, auth_token=None, process_name=None, user_name=None, code_location=None, get_code_version_fn=None, get_explicit_code_version_fn=None, init_context=True, do_report=None, report_to_file=None, offline_file_name=None, reporter=None, **kwargs):
        allow_reinit = kwargs["allow_reinit"] if "allow_reinit" in kwargs else False
        ksu_provided_inst = KensuProvider().instance()
        if ksu_provided_inst is not None and allow_reinit:
            logging.warning("KensuProvider.initKensu called more than once - reinitializing as requested by allow_reinit=True")
            KensuProvider().setKensu(None)
        if ksu_provided_inst is None or allow_reinit:
            from kensu.utils.kensu import Kensu
            pandas_support = kwargs["pandas_support"] if "pandas_support" in kwargs else True
            sklearn_support = kwargs["sklearn_support"] if "sklearn_support" in kwargs else True
            bigquery_support = kwargs["bigquery_support"] if "bigquery_support" in kwargs else None
            tensorflow_support = kwargs["tensorflow_support"] if "tensorflow_support" in kwargs else None
            bigquery_headers = kwargs["bigquery_headers"] if "bigquery_headers" in kwargs else None

            project_names = kwargs["project_names"] if "project_names" in kwargs else None
            environment = kwargs["environment"] if "environment" in kwargs else None
            timestamp = kwargs["timestamp"] if "timestamp" in kwargs else None
            logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None
            mapping = kwargs["mapping"] if "mapping" in kwargs else True
            report_in_mem = kwargs["report_in_mem"] if "report_in_mem" in kwargs else False
            get_code_version = kwargs["get_code_version"] if "get_code_version" in kwargs else None
            stats = kwargs["compute_stats"] if "compute_stats" in kwargs else True
            sql_util_url = kwargs["sql_util_url"] if "sql_util_url" in kwargs else None

            _kensu = Kensu(api_url=api_url, auth_token=auth_token, process_name=process_name, user_name=user_name,
                      code_location=code_location, init_context=init_context, do_report=do_report, pandas_support = pandas_support,
                      sklearn_support = sklearn_support, bigquery_support = bigquery_support, tensorflow_support = tensorflow_support, 
                      project_names=project_names,environment=environment,timestamp=timestamp,logical_naming=logical_naming,mapping=mapping, report_in_mem = report_in_mem,
                      report_to_file=report_to_file, offline_file_name=offline_file_name, reporter=reporter,
                      get_code_version=get_explicit_code_version_fn or get_code_version or get_code_version_fn,
                      compute_stats=stats, bigquery_headers = bigquery_headers, sql_util_url= sql_util_url)

            KensuProvider().setKensu(_kensu)
            return _kensu
        else:
            logging.error("Kensu default is already set kensu={}".format(str(ksu_provided_inst)))



import logging

from kensu.utils.dsl.extractors import ExtractorSupport
from kensu.utils.helpers import singleton

from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema


@singleton
class GenericDatasourceInfoSupport(ExtractorSupport):  # should extends some DamSupport class

    def is_supporting(self, df):
        return isinstance(df, KensuDatasourceAndSchema)

    def is_machine_learning(self, df):
        return False

    # return dict of doubles (stats)
    # stats are reported inside spark itself (?)
    def extract_stats(self, df):
        logging.debug('starting waiting for datastats for {}'.format(df.ksu_ds.name))
        stats = df.f_get_stats()
        logging.debug('done waiting for datastats for {}'.format(df.ksu_ds.name))
        return stats

    def extract_data_source(self,
                            ds,  # type: KensuDatasourceAndSchema
                            pl,
                            **kwargs):
        return ds.ksu_ds

    def extract_schema(
            self,
            data_source,
            df  # type: KensuDatasourceAndSchema
    ):
        return df.ksu_schema

    def extract_data_source_and_schema(self, df, pl, **kwargs):
        ds = self.extract_data_source(df, pl, **kwargs)
        sc = self.extract_schema(ds, df)
        return ds, sc



import logging
from hashlib import sha256

import pandas as pd
import google
from google.cloud.bigquery import Table
import datetime

import kensu
from kensu.client import *
from kensu.google.cloud.bigquery import Client
from kensu.google.cloud.bigquery.job.bigquery_stats import compute_bigquery_stats
from kensu.utils.dsl.extractors import ExtractorSupport
from kensu.utils.helpers import singleton, to_datasource
from kensu.utils.kensu_provider import KensuProvider


@singleton
class KensuBigQuerySupport(ExtractorSupport):  # should extends some KensuSupport class

    def is_supporting(self, table):
        return isinstance(table, google.cloud.bigquery.table.Table) or \
               isinstance(table, google.cloud.bigquery.table.RowIterator)

    def is_machine_learning(self, df):
        return False

    def skip_wr(self, df):
        if hasattr(df, 'ksu_dest'):
            return getattr(df, 'ksu_dest')
        return df

    # return list of FieldDef
    def extract_schema_fields(self, df):
        if isinstance(df, google.cloud.bigquery.table.Table) or isinstance(df, google.cloud.bigquery.table.RowIterator):
            return [FieldDef(name=str(k.name), field_type=k.field_type, nullable=k.is_nullable) for k in df.schema]


    def extract_location(self, df, location):
        # FIXME => what to do for RowIterator?
        if isinstance(df, google.cloud.bigquery.table.Table):
            return "bigquery:/" + df.path
        else:
            return location


    def extract_format(self, df, fmt):
        # FIXME => should do something about RowIterator...
        return "BigQuery Table"

    def tk(self, k, k1): return k + '.' + k1

    # return dict of doubles (stats)
    def extract_stats(self, df):
        # stats definitions are computed directly in wrapper in more efficient fashion, see `query.py`
        return None

    def extract_data_source(self, df, pl, **kwargs):
        # FIXME: is this not used?
        df=self.skip_wr(df)
        logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None

        location = self.extract_location(df, kwargs.get("location"))
        fmt = self.extract_format(df, kwargs.get("format"))

        if location is None or fmt is None:
            raise Exception(
                "cannot report new bigquery dataframe without location ({}) a format provided ({})!".format(location, fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)
        name = ('/').join(location.split('/')[-2:])
        return to_datasource(ds_pk=ds_pk, format=fmt, location=location, logical_naming=logical_naming, name=name)


    def extract_schema(self, data_source, df):
        df = self.skip_wr(df)
        fields = self.extract_schema_fields(df)
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:"+data_source.name, pk=sc_pk)
        return schema

    def extract_data_source_and_schema(self, df, pl, **kwargs):
        ds = self.extract_data_source(df, pl, **kwargs)
        sc = self.extract_schema(ds, df)
        return ds, sc


import itertools
from kensu.numpy import ndarray
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.helpers import eventually_report_in_mem
class chain(itertools.chain):
    deps = []

    @classmethod
    def from_iterable(self,iterables):
        # chain('ABC', 'DEF') --> A B C D E F
        for it in iterables:
            kensu = KensuProvider().instance()
            if isinstance(it,ndarray):
                orig_it = it
                it = it.get_nd()
                orig_ds = eventually_report_in_mem(
                    kensu.extractors.extract_data_source(orig_it, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, orig_it))
                self.deps.append(orig_sc)

            for element in it:
                yield element
            if self.deps != []:
                yield self.deps




class kensu_list(list):
    ksu_metadata = {}
    deps=[]
    def __init__(self, *args):
        list.__init__(self, *args)
        self.add_deps()

    def add_deps(self):
        from kensu.client.models.schema import Schema
        if isinstance(self[-1],list):
            if isinstance(self[-1][0],Schema):
                self.deps = self[-1]
                del self[-1]


import logging

import os
# disable Kensu collector when requested to do so (e.g. inside Apache Spark executor nodes)
if "KSU_DISABLE_PY_COLLECTOR" in os.environ:
    from numpy import *
else:

    from numpy import *
    import numpy as np
    from kensu.utils.kensu_provider import KensuProvider
    from kensu.utils.dsl import mapping_strategies
    from kensu.utils.helpers import eventually_report_in_mem, get_absolute_path
    from kensu.utils.wrappers import remove_ksu_wrappers


    class ndarrayDelegator(object):
        SKIP_KENSU_FIELDS = ["_ndarray__k_nd", "INTERCEPTORS", "_ksu_loc_id"]
        SKIP_KENSU_METHODS = ["get_nd", "kensu_init", "to_string"]


        def __getattribute__(self, name):
            attr_value = object.__getattribute__(self, name)
            if name == "_ndarray__k_nd":
                return attr_value
            elif name == "__class__":
                return attr_value
            elif name in ['mean','std']:
                # fixme: lin lost?
                attr_value = object.__getattribute__(self.get_nd(), name)
                return attr_value
            elif name in ['round','reshape']:
                attr_value = object.__getattribute__(self.get_nd(), name)
                return ndarrayDelegator.handle_callable(self, name, attr_value)
            elif hasattr(attr_value, '__call__'):
                #return ndarrayDelegator.handle_callable(self, name, attr_value)
                return attr_value
            else:
                return ndarrayDelegator.handle_field(self, name, attr_value)

        @staticmethod
        def handle_callable(kensu_nd, name, attr_value):
            nd = kensu_nd.get_nd()
            delegated_pd_attr = object.__getattribute__(nd, name)


            docstring = delegated_pd_attr.__doc__
            return ndarrayDelegator.create_function_wrapper(kensu_nd, name, docstring)

        @staticmethod
        def create_function_wrapper(kensu_nd, name, docstring):
            from kensu.pandas.data_frame import Series, DataFrame
            logging.debug("Kensu wrapping function name=" + name)

            def wrapper(*args, **kwargs):

                nd = kensu_nd.get_nd()
                df_attr = object.__getattribute__(nd, name)

                kensu = KensuProvider().instance()

                # set_axis updates the DataFrame directly (self) and returns None, we loose the initial dataframe
                # so we need to create the Kensu object before
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(kensu_nd, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, kensu_nd))

                new_args=remove_ksu_wrappers(args)
                new_args = tuple(new_args)
                result = df_attr(*new_args, **kwargs)
                original_result = result
                result = ndarrayDelegator.wrap_returned_df(kensu_nd, result, name, nd)

                #adding the datasources, schemas and lineage to the dependencies
                if (result is not None and (isinstance(result,DataFrame) or isinstance(result,ndarray))):
                    numpy_report(kensu_nd, result, name)


                if original_result is None:
                    return None
                else:
                    return result

            wrapper.__doc__ = docstring
            return wrapper

        @staticmethod
        def wrap_returned_df(kensu_s, returned, name, original):
            if isinstance(returned, np.ndarray) \
                    and not isinstance(returned, ndarray) \
                    and original is not returned:
                logging.debug("Wrapping attr access resulting ndarray in KensuDF for property name=" + name)
                tmp = ndarray.using(returned)
                returned = tmp
            return returned

        @staticmethod
        def handle_field(kensu_df, name, attr_value):
            # _DataFrame__k_df => points to DataFrame.__k_df
            if name in ndarrayDelegator.SKIP_KENSU_FIELDS:
                # return Kensu DF attr value right away
                logging.debug("Kensu attribute name=" + name)
                return attr_value
            else:
                logging.debug("NOT Kensu, let's delegate attribute name=" + name)

                nd = kensu_df.get_nd()
                nd_attr = object.__getattribute__(nd, name)
                result = ndarrayDelegator.wrap_returned_df(kensu_df, nd_attr, name, nd)

                kensu = KensuProvider().instance()
                # adding the datasources, schemas and lineage to the dependencies


                if result is not None and isinstance(result, ndarray):
                    if kensu.mapping:
                        numpy_report(kensu_df,result,name)



                    else:
                        orig_ds = kensu.extractors.extract_data_source(kensu_df, kensu.default_physical_location_ref)._report()
                        orig_sc = kensu.extractors.extract_schema(orig_ds, kensu_df)._report()

                        result_ds = kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref)._report()
                        result_sc = kensu.extractors.extract_schema(result_ds, result)._report()

                        kensu.add_dependency((nd, orig_ds, orig_sc), (result, result_ds, result_sc),
                                           mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

                return result


    class ndarray(ndarrayDelegator, np.ndarray):
        __k_nd = None

        # https://numpy.org/doc/stable/user/basics.subclassing.html
        def __new__(subtype, shape, dtype=float, buffer=None, offset=0, strides=None, order=None):
            obj = super(ndarray, subtype).__new__(subtype, shape, dtype, buffer, offset, strides, order)
            return obj

        def __getitem__(self, item):
            returned = self.get_nd()[item]
            # fixme: input lineage lost
            if isinstance(returned,str):
                return returned
            else:
                return ndarray.using(returned)

        def __repr__(self):
            nd = self.get_nd()
            return nd.__repr__()

        @staticmethod
        def remove_np_wrapper(other):
            if isinstance(other, ndarray):
                other_np = other.get_nd()
            else:
                other_np = other
            return other_np

        def wrapped_ndarray_binary_op(self, other_input, wrapped_fn, op_title=None):
            op_title = op_title or 'Numpy ' + str(wrapped_fn.__name__)
            input_kensu_nd = self
            other_np = ndarray.remove_np_wrapper(other_input)
            result = ndarray.using(wrapped_fn(other_np))
            numpy_report(input_kensu_nd, result, op_title)
            if isinstance(other_input, ndarray):
                numpy_report(other_input, result, op_title)
            return result

        def __sub__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__sub__)

        def __add__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__add__)

        def __mul__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__mul__)

        def __divmod__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__divmod__)

        def __truediv__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__truediv__)

        def __cmp__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__cmp__)

        def __eq__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__eq__)

        def __ne__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__ne__)

        def __lt__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__lt__)

        def __gt__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__gt__)

        def __le__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__le__)

        def __ge__(self, other):
            return self.wrapped_ndarray_binary_op(other, self.get_nd().__ge__)

        def __array_finalize__(self, obj):
            if obj is None: return
            self.__k_nd = getattr(obj, '__k_nd', None)

        @staticmethod
        def using(o):
            if isinstance(o, ndarray):
                d = ndarray.using(o.get_nd())
                return d
            else:
                d = ndarray(o.shape,o.dtype)
                d.kensu_init(o)
                return d

        def get_nd(self):
            # if the current Kensu ndarray isn't delegating... TODO.. need to have a type for this instead...
            if self.__k_nd is None:
                return self
            else:
                return self.__k_nd

        def kensu_init(self, d):
            self.__k_nd = d

    def wrap_save(method):
        def wrapper(*args, **kwargs):
            kensu = KensuProvider().instance()
            result = method(*args, **kwargs)


            loc = args[0]
            df = args[1]


            if df.__class__ == ndarray:

                location = get_absolute_path(loc)

                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,
                                                             logical_naming=kensu.logical_naming))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))


                result_ds = kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref, location = location,
                                                               logical_naming=kensu.logical_naming)._report()
                result_sc = kensu.extractors.extract_schema(result_ds,df)._report()

                kensu.real_schema_df[result_sc.to_guid()] = result
                

                if kensu.mapping == True:
                    for col in [s.name for s in result_sc.pk.fields]:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col),
                                                     "save")
                    kensu.report_with_mapping()


            return result

        wrapper.__doc__ = method.__doc__
        return wrapper

    savetxt = wrap_save(np.savetxt)


    def wrap_where(method):
        def wrapper(*args, **kwargs):
            kensu = KensuProvider().instance()

            origins = []
            new_args = []
            condition = [args[0]]
            for item in args:
                from kensu.pandas import Series
                if isinstance(item,Series):
                    new_args.append(item.get_s())
                else:
                    new_args.append(item)
            for item in args[1:]:
                if isinstance(item,Series):
                    origins.append(item)
            new_args = tuple(new_args)

            result = method(*new_args, **kwargs)
            kensu_result = ndarray.using(result)

            # Note: must pass a kensu wrapped result
            result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(kensu_result, kensu.default_physical_location_ref,
                                                           logical_naming=kensu.logical_naming))
            result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, kensu_result))

            for element in origins:
                df = element.get_s()

                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,
                                                                                      logical_naming=kensu.logical_naming))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))

                if kensu.mapping == True:
                    for col in [s.name for s in result_sc.pk.fields] :
                        for col_orig in [s.name for s in orig_sc.pk.fields]:

                            kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col_orig),
                                                         "Numpy Where")

            return kensu_result

        wrapper.__doc__ = method.__doc__
        return wrapper

    where = wrap_where(np.where)

    def wrap_unique(method):
        def wrapper(*args, **kwargs):

            kensu = KensuProvider().instance()

            new_args = remove_ksu_wrappers(args)
            new_args = tuple(new_args)

            result = method(*new_args, **kwargs)
            original_result = result
            # with extra kwargs unique sometimes return tuples...
            if isinstance(original_result, np.ndarray):
                ksu_result = ndarray.using(original_result)
                numpy_report(args, ksu_result, "Numpy unique")
            elif isinstance(original_result, tuple) and all([isinstance(x, np.ndarray) for x in original_result]):
                ksu_result = tuple([ndarray.using(x) for x in original_result])
                for ksu_result_item in ksu_result:
                    numpy_report(args, ksu_result_item, "Numpy unique")
            else:
                msg = "Kensu numpy.unique got an unexpected result which may be not fully tracked"
                logging.warning(msg)
                ksu_result = original_result
            return ksu_result

        wrapper.__doc__ = method.__doc__
        return wrapper

    unique = wrap_unique(np.unique)

    def wrap_abs(method):
        def wrapper(*args, **kwargs):
            new_args = remove_ksu_wrappers(args)
            new_args = tuple(new_args)

            original_result = method(*new_args, **kwargs)
            ksu_result = ndarray.using(original_result)
            numpy_report(args,ksu_result,"Numpy abs")

            return ksu_result

        wrapper.__doc__ = method.__doc__
        return wrapper

    abs = wrap_abs(np.abs)

    def wrap_round(method):
        def wrapper(*args, **kwargs):
            new_args = remove_ksu_wrappers(args)
            new_args = tuple(new_args)

            original_result = method(*new_args, **kwargs)
            ksu_result = ndarray.using(original_result)
            numpy_report(args,ksu_result,"Numpy abs")
            return ksu_result

        wrapper.__doc__ = method.__doc__
        return wrapper

    round = wrap_round(np.round)

    def wrap_concat(method):
        def wrapper(*args, **kwargs):
            new_args = remove_ksu_wrappers(args)
            new_args = tuple(new_args)

            original_result = method(*new_args, **kwargs)
            ksu_result = ndarray.using(original_result)
            for input_nd in args:
                numpy_report(input_nd,ksu_result,"Numpy concat")
            return ksu_result

        wrapper.__doc__ = method.__doc__
        return wrapper

    concatenate = wrap_concat(np.concatenate)


    def array(method):
        def wrapper(*args, **kwargs):
            args = list(args)
            obj = args[0]
            if isinstance(obj,list):
                args[0] = remove_ksu_wrappers(obj)
                new_args = args
            else:
                new_args = remove_ksu_wrappers(args)

            new_args = tuple(new_args)

            original_result = method(*new_args, **kwargs)
            ksu_result = ndarray.using(original_result)
            numpy_report(obj, ksu_result, "Numpy array")
            return ksu_result

        wrapper.__doc__ = method.__doc__
        return wrapper

    array = array(np.array)

    def extract_sc(kensu, nd):
        try:
            orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(nd, kensu.default_physical_location_ref,
                                                                                logical_naming=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, nd))
        except Exception:  # thrown by default extract_data_source/extract_schema
            msg = "Kensu numpy.array got an unexpected argument which can not be fully tracked (yet): " + str(type(nd))
            logging.warning(msg)
            #import traceback
            #traceback.print_stack()
            orig_sc = None

        return orig_sc


    def numpy_report(nd,result,name):
        orig_sc_list = []

        kensu = KensuProvider().instance()
        from kensu.itertools import kensu_list

        if isinstance(nd,kensu_list):
            orig_sc_list = nd.deps
        elif isinstance(nd, tuple) or isinstance(nd, list):
            # FIXME: fail tolerance for unexpected arguments
            for input_nd in nd:
                if isinstance(input_nd, kensu_list):
                    for x in input_nd.deps:
                        orig_sc_list.append(x)
                else:
                    orig_sc_list.append(extract_sc(kensu, input_nd))
        else:
            orig_sc_list.append(extract_sc(kensu, nd))

        result_ds = eventually_report_in_mem(
            kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref,
                                                 logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

        if kensu.mapping:
            for orig_sc in orig_sc_list:
                if orig_sc is not None:
                    for col in [s.name for s in result_sc.pk.fields]:
                        for col_orig in [s.name for s in orig_sc.pk.fields]:
                            kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col_orig),
                                                           name)


from kensu.gluonts.ksu_utils.dataset_helpers import make_dataset_reliable
import gluonts.model.deepar as dp
from gluonts.dataset.common import  ListDataset
from kensu.gluonts.mx.model.predictor import RepresentableBlockPredictor

class DeepAREstimator(dp.DeepAREstimator):

    def train(self, X):
        X, old_Field, dep_fields = make_dataset_reliable(X)
        if isinstance(X,ListDataset):
            print(X.list_data[0]['target'].__class__)

        result = super(DeepAREstimator,self).train(X)
        result.__class__ = RepresentableBlockPredictor
        if isinstance(X,ListDataset):
            X.list_data = old_Field
        return result

